# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-02 06:06
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [('auth', '__first__'), ('contenttypes', '__first__')]

    operations = [
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True,
                                        serialize=False, verbose_name='ID')),
                ('user_id', models.BigIntegerField()),
                ('object_id', models.PositiveIntegerField()),
                ('create_at', models.DateTimeField(auto_now_add=True)),
                ('content_type', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='contenttypes.ContentType')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='vote',
            unique_together=set([('user_id', 'content_type', 'object_id')]),
        ),
    ]



from django.db import models, transaction, IntegrityError
from django.db.models.query import QuerySet
from django.db.utils import OperationalError
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from vote.utils import instance_required, add_field_to_objects

UP = 0
DOWN = 1


class VotedQuerySet(QuerySet):

    """
    if call votes.annotate with an `user` argument then add `is_voted` to each
    instance.
    """

    def __init__(self, model=None, query=None, using=None, user_id=None,
                 hints=None):

        self.user_id = user_id
        super(VotedQuerySet, self).__init__(model, query, using, hints)

    def __iter__(self):
        super(VotedQuerySet, self).__iter__()

        if self.user_id is None:
            return iter(self._result_cache)

        objects = self._result_cache
        user_id = self.user_id

        objects = add_field_to_objects(self.model, objects, user_id)

        self._result_cache = objects
        return iter(objects)

    def _clone(self):
        c = super(VotedQuerySet, self)._clone()
        c.user_id = self.user_id
        return c


class _VotableManager(models.Manager):

    def __init__(self, through, model, instance, field_name='votes'):
        self.through = through
        self.model = model
        self.instance = instance
        self.field_name = field_name

    def vote(self, user_id, action):
        try:
            with transaction.atomic():
                self.instance = self.model.objects.select_for_update().get(
                    pk=self.instance.pk)

                content_type = ContentType.objects.get_for_model(self.model)
                try:
                    vote = self.through.objects.get(user_id=user_id,
                                                    content_type=content_type,
                                                    object_id=self.instance.pk)
                    if vote.action == action:
                        return False
                    vote.action = action
                    vote.save()

                    # will delete your up if you vote down some instance that
                    # you have vote up
                    voted_field = self.through.ACTION_FIELD.get(
                        int(not action))
                    setattr(self.instance, voted_field,
                            getattr(self.instance, voted_field) - 1)
                except self.through.DoesNotExist:
                    self.through.objects.create(user_id=user_id,
                                                content_type=content_type,
                                                object_id=self.instance.pk,
                                                action=action)

                statistics_field = self.through.ACTION_FIELD.get(action)
                setattr(self.instance, statistics_field,
                        getattr(self.instance, statistics_field) + 1)

                self.instance.save()

            return True
        except (OperationalError, IntegrityError):
            return False

    @instance_required
    def up(self, user_id):
        return self.vote(user_id, action=UP)

    @instance_required
    def down(self, user_id):
        return self.vote(user_id, action=DOWN)

    @instance_required
    def delete(self, user_id):
        try:
            with transaction.atomic():
                content_type = ContentType.objects.get_for_model(self.instance)

                try:
                    # select_for_update will add a write lock here
                    vote = self.through.objects.select_for_update().get(
                        user_id=user_id,
                        content_type_id=content_type.id,
                        object_id=self.instance.id
                    )
                except self.through.DoesNotExist:
                    return False

                self.instance = self.model.objects.select_for_update().get(
                    pk=self.instance.pk)
                statistics_field = self.through.ACTION_FIELD.get(vote.action)
                setattr(self.instance, statistics_field,
                        getattr(self.instance, statistics_field) - 1)

                self.instance.save()

                vote.delete()

            return True
        except (OperationalError, IntegrityError):
            # concurrent request may decrease num_vote field to negative
            return False

    @instance_required
    def get(self, user_id):
        return self.through.objects.filter(
            user_id=user_id,
            content_object=self.instance,
        ).first()

    @instance_required
    def exists(self, user_id, action=UP):
        return self.through.objects.filter(
            user_id=user_id,
            content_object=self.instance,
            action=action
        ).exists()

    def all(self, user_id, action=UP):
        content_type = ContentType.objects.get_for_model(self.model)

        object_ids = self.through.objects.filter(
            user_id=user_id,
            content_type=content_type,
            action=action).values_list('object_id', flat=True)

        return self.model.objects.filter(pk__in=list(object_ids))

    def count(self, action=UP):
        return self.through.votes_for(self.model,
                                      self.instance, action).count()

    def user_ids(self, action=UP):
        return self.through.votes_for(
            self.model, self.instance, action
        ).order_by('-create_at').values_list('user_id', 'create_at')

    def annotate(self, queryset=None, user_id=None,
                 reverse=True, sort=True):

        if queryset is not None:
            queryset = queryset
        else:
            queryset = self.model.objects.all()

        if sort:
            order = reverse and '-%s' % 'vote_score' or 'vote_score'
            queryset = queryset.order_by(order, '-id')

        return VotedQuerySet(model=queryset.model, query=queryset.query,
                             user_id=user_id)

    def vote_by(self, user_id, queryset=None, ids=None):
        if queryset is None and ids is None:
            raise ValueError("queryset or ids can not be None")

        if ids is not None:
            objects = self.model.objects.filter(id__in=ids)
            objects = sorted(objects, key=lambda x: ids.index(x.id))

            return add_field_to_objects(self.model, objects, user_id)
        else:
            return VotedQuerySet(model=queryset.model, query=queryset.query,
                                 user_id=user_id)


class VotableManager(GenericRelation):

    def __init__(self, through=None, manager=_VotableManager, **kwargs):
        from vote.models import Vote
        self.through = Vote if not through else through
        self.manager = manager
        kwargs['verbose_name'] = kwargs.get('verbose_name', _('Votes'))
        super(VotableManager, self).__init__(self.through, **kwargs)

    def __get__(self, instance, model):
        if instance is not None and instance.pk is None:
            raise ValueError(
                "%s objects need to have a primary key value "
                "before you can access their votes." % model.__name__)

        manager = self.manager(
            through=self.through,
            model=model,
            instance=instance,
            field_name=self.name,
        )

        return manager

    def contribute_to_class(self, cls, name):
        super(VotableManager, self).contribute_to_class(cls, name)
        setattr(cls, name, self)



from __future__ import absolute_import
from django.test import TestCase
from django.contrib.auth.models import User
from vote.models import Vote, UP, DOWN
from test.models import Comment


class VoteTest(TestCase):
    through = Vote
    model = Comment
    field_name = 'votes'

    def setUp(self):
        self.user1 = User.objects.create_user("test1", "test1@test.com",
                                              '111111')
        self.user2 = User.objects.create_user("test2", "test2@test.com",
                                              '111111')
        self.user3 = User.objects.create_user("test3", "test3@test.com",
                                              '111111')
        print('\n%s.%s.%s' % (self.__class__.__module__,
                              self.__class__.__name__,
                              self._testMethodName))

    def tearDown(self):
        self.model.objects.all().delete()
        self.through.objects.all().delete()
        User.objects.all().delete()

    def call_api(self, api_name, model=None, *args, **kwargs):
        model = model or self.model
        votes = getattr(model, self.field_name)
        api = getattr(votes, api_name)
        return api(*args, **kwargs)

    def test_vote_up(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")

        self.call_api('up', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, UP), 1)

        # call up again, will not increase count
        res = self.call_api('up', comment, self.user2.pk)
        self.assertEqual(res, False)
        self.assertEqual(self.call_api('count', comment, UP), 1)

    def test_vote_down(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")

        # no votes yet, no exception raised
        self.assertEqual(self.call_api('count', comment, DOWN), 0)
        self.call_api('down', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, DOWN), 1)

        self.call_api('up', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, UP), 1)
        self.call_api('down', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, DOWN), 1)

    def test_vote_delete(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")

        self.call_api('up', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, UP), 1)
        self.call_api('delete', comment, self.user2.pk)
        self.assertEqual(self.call_api('count', comment, UP), 0)

    def test_vote_exists(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")
        self.assertFalse(self.call_api('exists', comment, self.user2.pk,
                                       UP))
        self.call_api('up', comment, self.user2.pk)
        self.assertTrue(self.call_api('exists', comment,
                                      self.user2.pk, UP))

    def test_vote_get(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")
        self.assertIsNone(self.call_api('get', comment, self.user2.pk))
        self.call_api('up', comment, self.user2.pk)
        vote = Vote.objects.first()
        self.assertEqual(vote, self.call_api('get', comment, self.user2.pk))

    def test_vote_all(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")
        self.assertFalse(self.call_api('exists', comment, self.user2.pk))
        self.call_api('up', comment, self.user2.pk)
        self.assertEqual(len(self.call_api('all', comment, self.user2.pk)), 1)

    def test_vote_count(self):
        comment1 = self.model.objects.create(user_id=self.user1.pk,
                                             content="I'm a comment")
        comment2 = self.model.objects.create(user_id=self.user1.pk,
                                             content="I'm a comment too")
        self.call_api('up', comment1, self.user2.pk)
        self.call_api('up', comment2, self.user2.pk)
        self.assertEqual(self.call_api('count', comment1), 1)
        self.assertEqual(self.call_api('count', comment2), 1)
        self.assertEqual(self.call_api('count'), 2)

    def test_user_ids(self):
        comment = self.model.objects.create(user_id=self.user1.pk,
                                            content="I'm a comment")
        self.call_api('up', comment, self.user1.pk)
        self.assertEqual(len(list(self.call_api('user_ids', comment))), 1)

        self.call_api('up', comment, self.user2.pk)
        self.assertEqual(len(list(self.call_api('user_ids', comment))), 2)

    def test_vote_annotate(self):
        comments = [
            self.model(
                user_id=self.user1.pk,
                content="I'm a comment, sequence %s" % i) for i in range(10)
        ]

        self.model.objects.bulk_create(comments)
        comments = list(self.model.objects.all())

        comment1 = comments[0]
        self.call_api('up', comment1, self.user2.pk)

        comment2 = comments[1]
        self.call_api('up', comment2, self.user2.pk)
        self.call_api('up', comment2, self.user3.pk)

        # queryset=None, user_id=None
        comments = self.call_api('annotate')

        self.assertEqual(comments[0].pk, comment2.pk)
        self.assertEqual(comments[1].pk, comment1.pk)

        comments = list(self.call_api('annotate', reverse=False))
        self.assertEqual(comments[-1].pk, comment2.pk)
        self.assertEqual(comments[-2].pk, comment1.pk)

        for comment in comments:
            self.assertTrue(hasattr(comment, 'vote_score'))

        # call annotate with queryset and user_id
        comments = self.model.objects.filter(user_id=self.user1.pk)
        comments = self.call_api('annotate', queryset=comments,
                                 user_id=self.user2.pk)
        self.assertEqual(comments[0].pk, comment2.pk)
        self.assertEqual(comments[1].pk, comment1.pk)

        for comment in comments:
            self.assertTrue(hasattr(comment, 'vote_score'))
            self.assertTrue(hasattr(comment, 'is_voted_up'))
            self.assertTrue(hasattr(comment, 'is_voted_down'))
            if comment.pk in (comment1.pk, comment2.pk):
                self.assertTrue(comment.is_voted_up)
            else:
                self.assertFalse(comment.is_voted_up)

    def test_objects_with_status(self):
        comments = [
            self.model(
                user_id=self.user1.pk,
                content="I'm a comment, sequence %s" % i) for i in range(10)
        ]

        self.model.objects.bulk_create(comments)
        comments = list(self.model.objects.all())

        comment1 = comments[0]
        self.call_api('up', comment1, self.user2.pk)

        comment2 = comments[1]
        self.call_api('up', comment2, self.user2.pk)
        self.call_api('up', comment2, self.user3.pk)

        comment_ids = [comment2.id, comment1.id]

        votes = getattr(self.model, self.field_name)
        comments = votes.vote_by(self.user2.pk, ids=comment_ids)

        self.assertEqual(comments[0].id, comment2.id)
        for comment in comments:
            self.assertTrue(hasattr(comment, 'is_voted_up'))
            self.assertTrue(hasattr(comment, 'is_voted_down'))

        comment_ids = [comment1.id, comment2.id]

        votes = getattr(self.model, self.field_name)
        comments = votes.vote_by(self.user2.pk, ids=comment_ids)

        self.assertEqual(comments[0].id, comment1.id)
        for comment in comments:
            self.assertTrue(hasattr(comment, 'is_voted_up'))
            self.assertTrue(hasattr(comment, 'is_voted_down'))

        comments = votes.vote_by(self.user2.pk, queryset=Comment.objects.all())
        for comment in comments:
            self.assertTrue(hasattr(comment, 'is_voted_up'))
            self.assertTrue(hasattr(comment, 'is_voted_down'))

        self.assertRaises(ValueError, lambda: votes.vote_by(self.user2.pk))

    def test_vote_templatetag(self):
        comments = [
            self.model(
                user_id=self.user1.pk,
                content="I'm a comment, sequence %s" % i) for i in range(5)
        ]

        self.model.objects.bulk_create(comments)
        comments = list(self.model.objects.all())

        comment1 = comments[0]
        self.call_api('up', comment1, self.user1.pk)
        # test with anonymous user
        self.client.get('/comments/')
        # test with authenticated user
        self.client.login(username=self.user1.username, password='111111')
        self.client.get('/comments/')

    def test_votemixin_api(self):
        comment = self.model(
            user_id=self.user1.pk,
            content="I'm a comment")
        comment.save()
        self.assertEqual(comment.num_vote_up, 0)
        self.client.login(
            username=self.user1.username, password='111111')

        res = self.client.post('/api/comments/%s/vote/' % comment.id)
        self.assertEqual(res.status_code, 200)
        res = self.client.post('/api/comments/%s/vote/' % comment.id)
        self.assertEqual(res.status_code, 409)
        comment = Comment.objects.get(id=comment.id)
        self.assertEqual(comment.num_vote_up, 1)

        res = self.client.post('/api/comments/%s/vote/' %
                               comment.id, {"action": "down"})
        self.assertEqual(res.status_code, 200)
        comment = Comment.objects.get(id=comment.id)
        self.assertEqual(comment.num_vote_up, 0)
        self.assertEqual(comment.num_vote_down, 1)

        res = self.client.delete('/api/comments/%s/vote/' % comment.id)
        comment = Comment.objects.get(id=comment.id)
        self.assertEqual(res.status_code, 200)
        self.assertEqual(comment.num_vote_up, 0)
        self.assertEqual(comment.num_vote_down, 0)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2017-01-10 03:50
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0003_vote_action'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='vote',
            unique_together=set([('user_id', 'content_type',
                                  'object_id', 'action')]),
        ),
    ]



from __future__ import absolute_import

from django import template
from django.contrib.auth.models import AnonymousUser

from vote.models import UP

register = template.Library()


@register.simple_tag
def vote_exists(model, user=AnonymousUser(), action=UP):
    if user.is_anonymous:
        return False
    return model.votes.exists(user.pk, action=action)



from rest_framework.decorators import action
from rest_framework.views import Response

from vote.signals import post_voted


class VoteMixin:
    def get_instance(self, pk):
        return self.queryset.get(pk=pk)

    @action(detail=True, methods=('post', 'delete'))
    def vote(self, request, pk):
        obj = self.get_instance(pk)
        user_id = request.user.pk
        if request.method.lower() == 'post':
            action = request.data.get('action', 'up')
            voted = getattr(obj.votes, action)(user_id)
            if voted:
                post_voted.send(sender=self.queryset.model,
                                obj=obj,
                                user_id=user_id,
                                action=action)
            else:
                return Response(data={}, status=409)
        else:
            deleted = obj.votes.delete(user_id)
            if deleted:
                post_voted.send(sender=self.queryset.model,
                                obj=obj,
                                user_id=user_id,
                                action='delete')
        return Response({})



import os
from setuptools import setup, find_packages

import vote


with open('README.md') as f:
    README = f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-vote',
    version='.'.join(str(i) for i in vote.VERSION),
    packages=find_packages(exclude=('test*',)),
    include_package_data=True,
    license='BSD License',
    description='A simple Django app to conduct vote.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/shellfly/django-vote',
    author='shellfly',
    author_email='shell0fly@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from vote.managers import VotableManager

UP = 0
DOWN = 1


class VoteManager(models.Manager):

    def filter(self, *args, **kwargs):
        if 'content_object' in kwargs:
            content_object = kwargs.pop('content_object')
            content_type = ContentType.objects.get_for_model(content_object)
            kwargs.update({
                'content_type': content_type,
                'object_id': content_object.pk
            })

        return super(VoteManager, self).filter(*args, **kwargs)


class Vote(models.Model):
    ACTION_FIELD = {
        UP: 'num_vote_up',
        DOWN: 'num_vote_down'
    }

    user_id = models.BigIntegerField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()
    action = models.PositiveSmallIntegerField(default=UP)
    create_at = models.DateTimeField(auto_now_add=True)

    objects = VoteManager()

    class Meta:
        unique_together = ('user_id', 'content_type', 'object_id', 'action')
        index_together = ('content_type', 'object_id')

    @classmethod
    def votes_for(cls, model, instance=None, action=UP):
        ct = ContentType.objects.get_for_model(model)
        kwargs = {
            "content_type": ct,
            "action": action
        }
        if instance is not None:
            kwargs["object_id"] = instance.pk

        return cls.objects.filter(**kwargs)


class VoteModel(models.Model):
    vote_score = models.IntegerField(default=0, db_index=True)
    num_vote_up = models.PositiveIntegerField(default=0, db_index=True)
    num_vote_down = models.PositiveIntegerField(default=0, db_index=True)
    votes = VotableManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.vote_score = self.calculate_vote_score
        super(VoteModel, self).save(*args, **kwargs)

    @property
    def calculate_vote_score(self):
        return self.num_vote_up - self.num_vote_down

    @property
    def is_voted_up(self):
        try:
            return self._is_voted_up
        except AttributeError:
            return False

    @is_voted_up.setter
    def is_voted_up(self, value):
        self._is_voted_up = value

    @property
    def is_voted_down(self):
        try:
            return self._is_voted_down
        except AttributeError:
            return False

    @is_voted_down.setter
    def is_voted_down(self, value):
        self._is_voted_down = value



from functools import wraps
from django.contrib.contenttypes.models import ContentType


def instance_required(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        if self.instance is None:
            error = "Can't call %s with a non-instance manager" % func.__name__
            raise TypeError(error)

        return func(self, *args, **kwargs)

    return inner


def add_field_to_objects(model, objects, user_id):
    from vote.models import Vote, UP, DOWN
    content_type = ContentType.objects.get_for_model(model)
    object_ids = [r.id for r in objects]

    voted_object_ids = Vote.objects.filter(
        user_id=user_id,
        content_type=content_type,
        object_id__in=object_ids
    ).values_list("object_id", "action")

    for r in objects:
        r.is_voted_up = (r.pk, UP) in voted_object_ids
        r.is_voted_down = (r.pk, DOWN) in voted_object_ids

    return objects



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2017-01-06 07:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0002_auto_20161229_1022'),
    ]

    operations = [
        migrations.AddField(
            model_name='vote',
            name='action',
            field=models.PositiveSmallIntegerField(default=0),
        ),
    ]
